@@ -64,6 +64,10 @@ pub struct RawParser<'a> {
     containers: Vec<Container>,
     last_line_was_empty: bool,
 
+    /// In case we have a broken link/image reference, we can call this callback
+    /// with the reference name and use the link/title pair returned instead
+    broken_link_callback: Option<&'a Fn(&str) -> Option<(String, String)>>,
+
     // state for code fences
     fence_char: u8,
     fence_count: usize,
@@ -79,14 +83,19 @@ pub struct ParseInfo<'a> {
     pub links: HashMap<String, (Cow<'a, str>, Cow<'a, str>)>,
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq)]
 pub enum Tag<'a> {
     // block-level tags
     Paragraph,
     Rule,
+
+    /// A heading. The field indicates the level of the heading.
     Header(i32),
+
     BlockQuote,
     CodeBlock(Cow<'a, str>),
+
+    /// A list. If the list is ordered the field indicates the number of the first item.
     List(Option<usize>),  // TODO: add delim and tight for ast (not needed for html)
     Item,
     FootnoteDefinition(Cow<'a, str>),
@@ -101,11 +110,15 @@ pub enum Tag<'a> {
     Emphasis,
     Strong,
     Code,
+
+    /// A link. The first field is the destination URL, the second is a title
     Link(Cow<'a, str>, Cow<'a, str>),
+
+    /// An image. The first field is the destination URL, the second is a title
     Image(Cow<'a, str>, Cow<'a, str>),
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq)]
 pub enum Event<'a> {
     Start(Tag<'a>),
     End(Tag<'a>),
@@ -117,7 +130,7 @@ pub enum Event<'a> {
     HardBreak,
 }
 
-#[derive(Copy, Clone, Debug)]
+#[derive(Copy, Clone, Debug, PartialEq)]
 pub enum Alignment {
     None,
     Left,
@@ -126,18 +139,20 @@ pub enum Alignment {
 }
 
 bitflags! {
-    pub flags Options: u32 {
-        const OPTION_FIRST_PASS = 1 << 0,
-        const OPTION_ENABLE_TABLES = 1 << 1,
-        const OPTION_ENABLE_FOOTNOTES = 1 << 2,
+    pub struct Options: u32 {
+        const OPTION_FIRST_PASS = 1 << 0;
+        const OPTION_ENABLE_TABLES = 1 << 1;
+        const OPTION_ENABLE_FOOTNOTES = 1 << 2;
     }
 }
 
 const MAX_LINK_NEST: usize = 10;
 
 impl<'a> RawParser<'a> {
-    pub fn new_with_links(text: &'a str, opts: Options,
-            links: HashMap<String, (Cow<'a, str>, Cow<'a, str>)>) -> RawParser<'a> {
+    pub fn new_with_links_and_callback(text: &'a str, opts: Options,
+            links: HashMap<String, (Cow<'a, str>, Cow<'a, str>)>,
+            callback: Option<&'a Fn(&str) -> Option<(String, String)>>)
+    -> RawParser<'a> {
         let mut ret = RawParser {
             text: text,
             off: if text.starts_with("\u{FEFF}") { 3 } else { 0 },
@@ -153,6 +168,8 @@ impl<'a> RawParser<'a> {
             fence_count: 0,
             fence_indent: 0,
 
+            broken_link_callback: callback,
+
             // info, used in second pass
             loose_lists: HashSet::new(),
             links: links,
@@ -162,8 +179,13 @@ impl<'a> RawParser<'a> {
         ret
     }
 
+    pub fn new_with_links(text: &'a str, opts: Options,
+            links: HashMap<String, (Cow<'a, str>, Cow<'a, str>)>) -> RawParser<'a> {
+        Self::new_with_links_and_callback(text, opts, links, None)
+    }
+
     pub fn new(text: &'a str, opts: Options) -> RawParser<'a> {
-        RawParser::new_with_links(text, opts, HashMap::new())
+        Self::new_with_links(text, opts, HashMap::new())
     }
 
     // offset into text representing current parse position, hopefully
@@ -356,29 +378,24 @@ impl<'a> RawParser<'a> {
             if at_eol {
                 self.off += n;
                 self.state = State::StartBlock;
-                // two empty lines closes lists and footnotes
+                // two empty lines closes lists, one empty line closes a footnote
                 let (n, empty_lines) = self.scan_empty_lines(&self.text[self.off ..]);
-                //println!("{} empty lines (n = {})", empty_lines, n);
-                let mut closed = false;
-                if empty_lines >= 1 {
-                    let mut close_tags: Vec<&mut (Tag<'a>, usize, usize)> = self.stack.iter_mut().skip_while(|tag| {
-                        match tag.0 {
-                            Tag::List(_) | Tag::FootnoteDefinition(_) => false,
-                            _ => true,
-                        }
-                    }).collect();
-                    if !close_tags.is_empty() {
-                        for tag in &mut close_tags {
+                for i in (0..self.stack.len()).rev() {
+                    let is_break = match self.stack[i].0 {
+                        Tag::List(_) => empty_lines >= 1,
+                        Tag::Item => false,
+                        Tag::FootnoteDefinition(_) => true,
+                        _ => break,
+                    };
+                    if is_break {
+                        for tag in &mut self.stack[i..] {
                             tag.1 = self.off; // limit
                             tag.2 = self.off; // next
                         }
-                        close_tags[0].2 = self.off + n; // next
-                        closed = true;
+                        self.stack[i].2 = self.off + n; // next
+                        return Some(self.end());
                     }
                 }
-                if closed {
-                    return Some(self.end());
-                }
                 self.off += n;
                 if let Some(_) = self.at_list(2) {
                     self.last_line_was_empty = true;
@@ -700,10 +717,6 @@ impl<'a> RawParser<'a> {
                     self.state = State::CodeLineStart;
                     break;
                 }
-                b'\t' => {
-                    if i > beg { break; }
-                    return self.char_tab();
-                }
                 b'\r' => {
                     // just skip it (does not support '\r' only line break)
                     if i > beg { break; }
@@ -749,11 +762,34 @@ impl<'a> RawParser<'a> {
                 data.starts_with("<!")
     }
 
+    // http://spec.commonmark.org/0.26/#html-blocks
     fn get_html_tag(&self) -> Option<&'static str> {
-        static TAGS: &'static [&'static str; 3] = &["</script>", "</pre>", "</style>"];
-        for tag in TAGS {
-            if self.text[self.off + 1..].starts_with(&tag[2..]) {
-                return Some(tag);
+        static BEGIN_TAGS: &'static [&'static str; 3] = &["script", "pre", "style"];
+        static END_TAGS: &'static [&'static str; 3] = &["</script>", "</pre>", "</style>"];
+
+        for (beg_tag, end_tag) in BEGIN_TAGS.iter().zip(END_TAGS.iter()) {
+            if self.off + 1 + beg_tag.len() < self.text.len() &&
+               self.text[self.off + 1..].starts_with(&beg_tag[..]) {
+                let pos = self.off + beg_tag.len() + 1;
+                let s = self.text.as_bytes()[pos];
+                if s == b' ' || s == b'\n' || s == b'>' {
+                    return Some(end_tag);
+                }
+            }
+        }
+        static ST_BEGIN_TAGS: &'static [&'static str; 3] = &["<!--", "<?", "<![CDATA["];
+        static ST_END_TAGS: &'static [&'static str; 3] = &["-->", "?>", "]]>"];
+        for (beg_tag, end_tag) in ST_BEGIN_TAGS.iter().zip(ST_END_TAGS.iter()) {
+            if self.off + 1 + beg_tag.len() < self.text.len() &&
+               self.text[self.off + 1..].starts_with(&beg_tag[..]) {
+                return Some(end_tag);
+            }
+        }
+        if self.off + 4 < self.text.len() &&
+           self.text[self.off + 1..].starts_with("<!") {
+            let c = self.text[self.off + 4..self.off + 5].chars().next().unwrap();
+            if c >= 'A' && c <= 'Z' {
+                return Some(">");
             }
         }
         None
@@ -1275,7 +1311,17 @@ impl<'a> RawParser<'a> {
             let reference = self.normalize_link_ref(&data[ref_beg..ref_end]);
             let (dest, title) = match self.links.get(&reference) {
                 Some(&(ref dest, ref title)) => (dest.clone(), title.clone()),
-                None => return None
+                None => {
+                    if let Some(ref callback) = self.broken_link_callback {
+                        if let Some(val) = callback(&reference) {
+                            (val.0.into(), val.1.into())
+                        } else {
+                            return None;
+                        }
+                    } else {
+                        return None;
+                    }
+                }
             };
             (dest, title, text_beg, text_end, i)
         };
@@ -1344,7 +1390,14 @@ impl<'a> RawParser<'a> {
                 None
             } else {
                 let space = scan_whitespace_no_nl(&data[len + n_colon..]);
-                Some((name, len + n_colon + space))
+                // skip newline if definition is on a line by itself, as likely that
+                // means the footnote definition is a complex block.
+                let mut i = len + n_colon + space;
+                if let (n, true) = scan_eol(&data[i..]) {
+                    let (n_containers, _, _) = self.scan_containers(&data[i + n ..]);
+                    i += n + n_containers;
+                }
+                Some((name, i))
             }
         })
     }
@@ -1529,7 +1582,11 @@ impl<'a> RawParser<'a> {
         while i < data.len() {
             match data.as_bytes()[i] {
                 b'>' => return i + 1,
-                b'\n' => i += self.scan_whitespace_inline(&data[i..]),
+                b'\n' => {
+                    let n = self.scan_whitespace_inline(&data[i..]);
+                    if n == 0 { break; }
+                    i += n;
+                }
                 _ => i += 1
             }
         }
