@@ -1,20 +1,26 @@
 // Copyright 2015 Google Inc. All rights reserved.
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
+// Permission is hereby granted, free of charge, to any person obtaining a copy
+// of this software and associated documentation files (the "Software"), to deal
+// in the Software without restriction, including without limitation the rights
+// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+// copies of the Software, and to permit persons to whom the Software is
+// furnished to do so, subject to the following conditions:
 //
-//     http://www.apache.org/licenses/LICENSE-2.0
+// The above copyright notice and this permission notice shall be included in
+// all copies or substantial portions of the Software.
 //
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+// THE SOFTWARE.
 
 //! Main public pull parse interface, running two passes over input.
 
-use parse::{RawParser, Event, Tag, Options};
+use parse::{RawParser, Event, Tag, Options, OPTION_FIRST_PASS};
 use std::collections::HashSet;
 
 pub struct Parser<'a> {
@@ -25,17 +31,30 @@ pub struct Parser<'a> {
 
 impl<'a> Parser<'a> {
     pub fn new(text: &'a str) -> Parser<'a> {
+        Parser::new_ext(text, Options::empty())
+    }
+    pub fn new_ext(text: &'a str, opts: Options) -> Parser<'a> {
+        Self::new_with_broken_link_callback(text, opts, None)
+    }
+
+    /// In case the parser encounters any potential links that have a broken
+    /// reference (e.g `[foo]` when there is no `[foo]: ` entry at the bottom)
+    /// the provided callback will be called with the reference name,
+    /// and the returned pair will be used as the link name and title if not
+    /// None.
+    pub fn new_with_broken_link_callback(text: &'a str, mut opts: Options,
+            callback: Option<&'a Fn(&str) -> Option<(String, String)>>)
+            -> Parser<'a>  {
+        opts.remove(OPTION_FIRST_PASS);
         // first pass, collecting info
-        let mut opts = Options::new();
-        opts.set_first_pass();
-        let mut first = RawParser::new(text, opts);
+        let first_opts = opts | OPTION_FIRST_PASS;
+        let mut first = RawParser::new(text, first_opts);
         while first.next().is_some() { }
         let info = first.get_info();
 
         // second pass
         let loose_lists = info.loose_lists;
-        let opts = Options::new();
-        let second = RawParser::new_with_links(text, opts, info.links);
+        let second = RawParser::new_with_links_and_callback(text, opts, info.links, callback);
 
         //println!("loose lists: {:?}", info.loose_lists);
         Parser {
